#--------------------------------------------------------------------------------
# 참조 모듈 목록.
#--------------------------------------------------------------------------------
from __future__ import annotations
from typing import Awaitable, Callable, Final, Generic, Iterable, Iterator, Optional, Sequence, Type, TypeVar, Tuple, Union
from typing import ItemsView, KeysView, ValuesView
from typing import IO, TextIO, BinaryIO
from typing import Any, List, Dict, Set
from typing import cast, overload
import builtins


#--------------------------------------------------------------------------------
# 패키지 포함 목록.
#--------------------------------------------------------------------------------
from .directory import Directory
from .drive import Drive
from .entry import Entry
from .file import File
from .stroage import Storage


#--------------------------------------------------------------------------------
# 참조 모듈 목록.
#--------------------------------------------------------------------------------
from __future__ import annotations
from typing import Awaitable, Callable, Final, Generic, Iterable, Iterator, Optional, Sequence, Type, TypeVar, Tuple, Union
from typing import ItemsView, KeysView, ValuesView
from typing import IO, TextIO, BinaryIO
from typing import Any, List, Dict, Set
from typing import cast, overload
import builtins


#--------------------------------------------------------------------------------
# 파이썬 빌트인 함수들을 바탕으로 리플렉션 기능으로 확장한 클래스.
#--------------------------------------------------------------------------------
class Reflection:
	"""파이썬 리플렉션 클래스"""
	#--------------------------------------------------------------------------------
	# 대상이 타입 인스턴스인지 여부 반환.
	# - instance = Hello()
	# - instance: 인스턴스
	# - instance.__class__: 인스턴스의 클래스 타입 (class Hello)
	# - instance.__class__ == Hello: 동일
	#
	# - instance.__class__.__class__: 클래스의 클래스 타입 (class type)
	# - instance.__class__.__class__ == type: 동일
	# - 타입은 "클래스의 메타클래스"로서 클래스는 타입의 인스턴스라고 할 수 있다.
	# - instance < class < type
	# - instance는 type과 직접적인 관련이 없기 때문에 같지 않다.
	#--------------------------------------------------------------------------------
	@staticmethod
	def IsType(target: object) -> bool:
		"""대상이 타입 클래스인지 여부 반환."""
		if not builtins.isinstance(target, builtins.type):
			return False
		return True


	#--------------------------------------------------------------------------------
	# 대상이 클래스 타입에 해당하는 인스턴스인지 여부 반환.
	#--------------------------------------------------------------------------------
	@staticmethod
	def IsInstanceType(target: object, classType: Type) -> bool:
		"""대상의 타입이 지정 타입과 동일한지 여부 반환."""
		if not builtins.isinstance(target, classType):
			return False
		return True


	#--------------------------------------------------------------------------------
	# 대상이 클래스 타입 목록 중 하나에 해당하는 인스턴스인지 여부 반환.
	#--------------------------------------------------------------------------------
	@staticmethod
	def IsInstanceInTypes(target: object, classTypes: List[Type]) -> bool:
		"""대상이 타입 중 하나에 속해있는지 여부 반환."""
		if not builtins.isinstance(target, tuple(classTypes)):
			return False
		return True


	#--------------------------------------------------------------------------------
	# 어트리뷰트의 값을 설정. (혹은 생성)
	#--------------------------------------------------------------------------------
	@staticmethod
	def SetAttributeValue(target: object, attributeName: str, value: Any, allowCreation: bool = True) -> bool:
		"""대상의 어트리뷰트를 설정. 어트리뷰트가 없다면 새로 추가."""
		realAttributeName = Reflection.GetAttributeName(target, attributeName)
		if realAttributeName:
			builtins.setattr(target, attributeName, value)
			return True
		elif allowCreation:
			builtins.setattr(target, attributeName, value)
			return True
		else:
			return False

	#--------------------------------------------------------------------------------
	# 어트리뷰트 생성.
	#--------------------------------------------------------------------------------
	@staticmethod
	def CreateAttribute(target: object, attributeName: str, attributeValue: Any) -> bool:
		"""대상에게 어트리뷰트를 추가."""
		realAttributeName = Reflection.GetAttributeName(target, attributeName)
		if realAttributeName:
			return False	
		builtins.setattr(target, attributeName, attributeValue)
		return True
	

	#--------------------------------------------------------------------------------
	# 어트리뷰트 제거.
	#--------------------------------------------------------------------------------
	@staticmethod
	def DeleteAttribute(target: object, attributeName: str) -> bool:
		"""대상의 어트리뷰트를 제거."""
		attributeName = Reflection.GetAttributeName(target, attributeName)
		if not attributeName:
			return False	
		builtins.delattr(attributeName)
		return True


	#--------------------------------------------------------------------------------
	# 실제 어트리뷰트 이름 반환.
	# - 입력값에서 대소문자가 다르더라도 실제 어트리뷰트이름이 있다면 반환.
	# - 존재한다면 실제 어트리뷰트 이름을 반환. (입력값과 동일할 수도 다를 수도 있음)
	# - 존재하지 않으면 빈문자열 반환.
	#--------------------------------------------------------------------------------
	@staticmethod
	def GetAttributeName(target: object, attributeName: str) -> str:
		"""대상에게 어트리뷰트가 존재하면 그 이름을 반환하고 없으면 빈문자열 반환. (입력값과 같을 수도 다를 수도 있음)"""
		if not target:
			return str()
		if not attributeName:
			return str()
		attributeNameLower: str = attributeName.lower()
		for attributeName in builtins.dir(target):
			realAttributeNameLower: str = attributeName.lower()
			if attributeNameLower == realAttributeNameLower:
				return attributeName				
		return str()


	#--------------------------------------------------------------------------------
	# 어트리뷰트의 값을 반환.
	#--------------------------------------------------------------------------------
	@staticmethod
	def GetAttributeValue(target: object, attributeName: str, defaultValue: Any = None) -> Any:
		"""대상의 어트리뷰트를 반환"""
		if not target:
			return defaultValue
		if not attributeName:
			return defaultValue
		attributeName: str = Reflection.GetAttributeName(target, attributeName)
		if not attributeName:
			return defaultValue
		attributeValue: Any = builtins.getattr(target, attributeName)
		return attributeValue

	#--------------------------------------------------------------------------------
	# 어트리뷰트의 타입을 반환.
	# - 값으로부터 추론하므로 값이 없을 경우는 NoneType이다.
	# - 파이썬은 동적타이핑 언어라서 이렇게 되었을 경우 얻어올 수 있는 방법은 없다.
	# - 파이랜스 같은 인텔리센스는 타입힌팅을 통해 코드에디터레벨에서 정적타이핑인 척만 할 뿐이다.
	#--------------------------------------------------------------------------------
	@staticmethod
	def GetAttributeType(target: object, attributeName: str) -> type:
		"""대상이 어트리뷰트를 가지고 있는지 여부 반환"""
		if not Reflection.HasAttribute(attributeName):
			return None
		attributeValue = Reflection.GetAttributeValue(target, attributeName)
		return type(attributeValue)
	

	#--------------------------------------------------------------------------------
	# 어트리뷰트의 존재 유무 판단.
	#--------------------------------------------------------------------------------
	@staticmethod
	def HasAttribute(target: object, attributeName: str) -> bool:
		"""대상이 어트리뷰트를 가지고 있는지 여부 반환"""
		if not target:
			return False
		if not attributeName:
			return False
		attributeName: str = Reflection.GetAttributeName(target, attributeName)
		if not attributeName:
			return False
		return True


#--------------------------------------------------------------------------------
# 참조 모듈 목록.
#--------------------------------------------------------------------------------
from __future__ import annotations
from typing import Awaitable, Callable, Final, Generic, Iterable, Iterator, Optional, Sequence, Type, TypeVar, Tuple, Union
from typing import ItemsView, KeysView, ValuesView
from typing import IO, TextIO, BinaryIO
from typing import Any, List, Dict, Set
from typing import cast, overload
import builtins
import os
from .filter import Filter


#--------------------------------------------------------------------------------
# 파일 필터 클래스.
# - 루트 디렉토리를 넣고 for문을 돌리면 조건에 부합하는 파일경로만 출력됨.
#--------------------------------------------------------------------------------
class FileFilter(Filter[str]):
	def __init__(thisInstance, rootPath: str, predicate: Callable[[str], bool] = None):
		base = super()
		base.__init__()

		# 제너레이터 표현식. (Iterable 호출시 사용되어 메모리 할당됨)
		thisInstance.Iterable = (os.path.join(root, file) for root, directories, files in os.walk(rootPath) for file in files)
		thisInstance.Predicate = predicate


#--------------------------------------------------------------------------------
# 참조 모듈 목록.
#--------------------------------------------------------------------------------
from __future__ import annotations
from typing import Awaitable, Callable, Final, Generic, Iterable, Iterator, Optional, Sequence, Type, TypeVar, Tuple, Union
from typing import ItemsView, KeysView, ValuesView
from typing import IO, TextIO, BinaryIO
from typing import Any, List, Dict, Set
from typing import cast, overload
import builtins


#--------------------------------------------------------------------------------
# 패키지 포함 목록.
#--------------------------------------------------------------------------------
from .environment import Environment
from .exitcodetype import ExitCodeType
from .path import Path
from .platformtype import PlatformType


#--------------------------------------------------------------------------------
# 참조 모듈 목록.
#--------------------------------------------------------------------------------
from __future__ import annotations
from typing import Awaitable, Callable, Final, Generic, Iterable, Iterator, Optional, Sequence, Type, TypeVar, Tuple, Union
from typing import ItemsView, KeysView, ValuesView
from typing import Any, List, Dict, Set
from typing import cast, overload
from enum import Enum, auto


#--------------------------------------------------------------------------------
# 플랫폼 타입.
#--------------------------------------------------------------------------------
class PlatformType(Enum):
	UNKNOWN = auto()
	
	# windows.
	WINDOWS = auto()

	# linux.
	LINUX = auto()
	ANDROID = auto()
	STEAMOS = auto()

	# apple.
	MACOS = auto()
	IOS = auto()
	IPADOS = auto()
	WATCHOS = auto()
	TVOS = auto()
	VISIONOS = auto()

	# web.
	WEBGL = auto()
	WEBGPU = auto()


#--------------------------------------------------------------------------------
# 참조 모듈 목록.
#--------------------------------------------------------------------------------
from __future__ import annotations
from typing import Awaitable, Callable, Final, Generic, Iterable, Iterator, Optional, Sequence, Type, TypeVar, Tuple, Union
from typing import ItemsView, KeysView, ValuesView
from typing import IO, TextIO, BinaryIO
from typing import Any, List, Dict, Set
from typing import cast, overload
import builtins


#--------------------------------------------------------------------------------
# 패키지 포함 목록.
#--------------------------------------------------------------------------------
from .console import Console
from .container import Dictionary
from .core import AnonymousObject, UnnamedClass
from .core import AsyncEventHandler
from .core import BaseClass
from .core import BaseConstant, Constant
from .core import BaseMetaClass, MetaClass, Meta
from .core import BaseNode, Node
from .core import EnumFlag, auto
from .core import EventHandler
from .core import Interface, InterfaceMetaClass, abstractmethod, IInterface, IInterfaceMetaClass, NoInstantiationMetaClass
from .core import ManagedObject, ManagedObjectGarbageCollection
from .core import Reflection
from .core import WeakedReference
from .dataformat import Document, Element, Path
from .decorator import classproperty, overridemethod
from .ecs import System
from .ecs import Component, Component
from .ecs import Entity, Entity
from .environment import Environment, ExitCodeType, Path, PlatformType
from .exception import SingletonError
from .filesystem import Entry, Directory, Drive, File, Storage
# from .future import EventNode, Node
# from .generic import *
from .http import HTTPStatusError
from .management import BaseManager, BaseRepository, Repository, SharedClass, Singleton
from .task import Target, Task, TaskRunner
from .utility import Filter, FileFilter, Logger, LogLevel, JSONUtility, StringUtility


#--------------------------------------------------------------------------------
# 참조 모듈 목록.
#--------------------------------------------------------------------------------
from __future__ import annotations
from typing import Awaitable, Callable, Final, Generic, Iterable, Iterator, Optional, Sequence, Type, TypeVar, Tuple, Union
from typing import ItemsView, KeysView, ValuesView
from typing import IO, TextIO, BinaryIO
from typing import Any, List, Dict, Set
from typing import cast, overload
import builtins


#--------------------------------------------------------------------------------
# 패키지 포함 목록.
#--------------------------------------------------------------------------------
from .anonymousclass import AnonymousObject, UnnamedClass
from .asynceventhandler import AsyncEventHandler
from .baseclass import BaseClass
# from .baseclass import BaseClass as Object
from .baseconstant import BaseConstant
from .baseconstant import BaseConstant as Constant
from .basemetaclass import BaseMetaClass
from .basemetaclass import BaseMetaClass as MetaClass
from .basemetaclass import BaseMetaClass as Meta
from .basenode import BaseNode
from .basenode import BaseNode as Node
from .enumflag import EnumFlag, auto
from .eventhandler import EventHandler
from .interface import Interface, InterfaceMetaClass, abstractmethod, IInterface, IInterfaceMetaClass, NoInstantiationMetaClass
from .managedobject import ManagedObject, ManagedObjectGarbageCollection
from .reflection import Reflection
from .weakedreference import WeakedReference


#--------------------------------------------------------------------------------
# 참조 모듈 목록.
#--------------------------------------------------------------------------------
from __future__ import annotations
from typing import Awaitable, Callable, Final, Generic, Iterable, Iterator, Optional, Sequence, Type, TypeVar, Tuple, Union
from typing import ItemsView, KeysView, ValuesView
from typing import IO, TextIO, BinaryIO
from typing import Any, List, Dict, Set
from typing import cast, overload
import builtins
from abc import ABC as Interface
from abc import ABC as IInterface
from abc import ABCMeta as InterfaceMetaClass
from abc import ABCMeta as IInterfaceMetaClass
from abc import abstractmethod
from .basemetaclass import BaseMetaClass


#--------------------------------------------------------------------------------
# 결합된 인터페이스 메타 클래스.
# - 인터페이스를 상속받은 클래스는 인터페이스메타클래스를 사용하기 때문에 메타클래스의 변형을 다룰때는 다음과 같이 다룬다.
# - 이렇게 두개의 메타를 같이 받고, 아래 주석처럼 다중상속을 통해 결합된 메타클래스를 만들고 해당 내용에 로직을 구현하고 메타클래스로 사용하면 된다.
# - class INode(Interface, metaclass = CombinedInterfaceMetaClass) 이런식으로 인터페이스를 선언하는 것이다.
#--------------------------------------------------------------------------------
class CombinedInterfaceMetaClass(BaseMetaClass, InterfaceMetaClass):
	pass


#--------------------------------------------------------------------------------
# 인스턴스화를 허용하지 않는 메타 클래스.
# - 샘플 겸해서 만들어 본 인터페이스의 상속 클래스를 외부에서 함수 방식의 호출로 객체가 생성되지 않도록 해주는 인터페이스용 메타 클래스.
#--------------------------------------------------------------------------------
class NoInstantiationMetaClass (BaseMetaClass, InterfaceMetaClass):
	#--------------------------------------------------------------------------------
	# 호출 프로토콜 메서드.
	# - 이 예제는 대상 클래스 객체가 생성될 때 익셉션을 라이즈한다.
	# - 쉽게 말해 a = A() 일때, A()는 __call__ 로 들어오니 이를 막는 것이다.
	# - 직접적인 생성자 호출만 막는 것이므로 다음과 같이 a = Object.__new__(A) 를 통해 클래스를 넘겨 객체를 생성할 수 있기 때문에 어디까지나 쓰지말라는 경고 차원이다.
	# - 파이썬은 견고하지 않기 때문에 사실 필요에 따라 얼마든 멋대로 생성해 낼 수 있다.
	# - 그럼에도 불구하고 사용해야하는 이유는 명시적으로 하지 않도록 설정하는 것이 안정적인 결과물을 만들어내는데 보탬이 되기 때문이다.
	# - 주석으로 경고만 하고 문법적으로 허용해버리면 안읽어서 몰랐다고 하면 그만이다. 힘들게 막아놔야 힘들게 뚫는 코드가 올라오게 되니 상대의 의도가 보다 명확해진다. 
	#--------------------------------------------------------------------------------
	def __call__(thisClassType, *argumentList, **argumentDictionary) -> None:
		raise ValueError("This object must not be created directly via its constructor.")


#--------------------------------------------------------------------------------
# 참조 모듈 목록.
#--------------------------------------------------------------------------------
from __future__ import annotations
from typing import Awaitable, Callable, Final, Generic, Iterable, Iterator, Optional, Sequence, Type, TypeVar, Tuple, Union
from typing import ItemsView, KeysView, ValuesView
from typing import IO, TextIO, BinaryIO
from typing import Any, List, Dict, Set
from typing import cast, overload
import builtins


#--------------------------------------------------------------------------------
# 패키지 포함 목록.
#--------------------------------------------------------------------------------
from .filter import Filter
from .filefilter import FileFilter
from .jsonutility import JSONUtility
from .logger import Logger
from .loglevel import LogLevel
from .stringutility import StringUtility
from .xmlutility import XMLUtility


#--------------------------------------------------------------------------------
# 참조 모듈 목록.
#--------------------------------------------------------------------------------
from __future__ import annotations
from typing import Awaitable, Callable, Final, Generic, Iterable, Iterator, Optional, Sequence, Type, TypeVar, Tuple, Union
from typing import ItemsView, KeysView, ValuesView
from typing import IO, TextIO, BinaryIO
from typing import Any, List, Dict, Set
from typing import cast, overload
import builtins
from enum import Enum
import logging
from logging import Logger as PythonStandardLogger
from logging import StreamHandler, FileHandler, Formatter
from logging.handlers import QueueHandler, QueueListener
import os
from queue import Queue
# from asyncio import Queue
from ..environment import Path
from .loglevel import LogLevel



#--------------------------------------------------------------------------------
# 로그 클래스.
#--------------------------------------------------------------------------------
class Logger:
	#--------------------------------------------------------------------------------
	# 멤버 변수 목록.
	#--------------------------------------------------------------------------------
	__name: str
	__filePath: str
	__level: LogLevel
	__logger: PythonStandardLogger
	__consoleHandler: StreamHandler
	__fileHandler: FileHandler
	__formatter: Formatter
	# __queue: Queue
	# __queueHandler: QueueHandler
	# __queueListener: QueueListener


	#--------------------------------------------------------------------------------
	# 로거 이름.
	#--------------------------------------------------------------------------------
	@property
	def Name(thisInstance) -> str:
		return thisInstance.__name
	

	#--------------------------------------------------------------------------------
	# 로그 파일 위치.
	#--------------------------------------------------------------------------------
	@property
	def FilePath(thisInstance) -> str:
		return thisInstance.__filePath


	#--------------------------------------------------------------------------------
	# 로그 수준.
	#--------------------------------------------------------------------------------
	@property
	def Level(thisInstance) -> LogLevel:
		return thisInstance.__level


	#--------------------------------------------------------------------------------
	# 프로세스 아이디.
	#--------------------------------------------------------------------------------
	@property
	def ProcessID(thisInstance) -> int:
		return os.getpid()
	

	#--------------------------------------------------------------------------------
	# 초기화 라이프사이클 메서드.
	# - 로그파일을 None 을 입력하면 
	#--------------------------------------------------------------------------------
	def __init__(thisInstance, name: str, filePath: str, level: LogLevel = LogLevel.DEBUG, 
			  useShowProcessID: bool = False, useShowName: bool = False) -> None:
		thisInstance.__name = name
		thisInstance.__filePath = filePath
		thisInstance.__level = level
	
		# 이름이 없다면 파일명에서 이름을 가져옴.
		if not name:
			if not filePath:
				raise NameError(name)
			else:
				_, fileName = Path.GetPathAndFileNameFromFileFullPath(filePath)
				name = os.path.splitext(fileName)

		thisInstance.__logger: PythonStandardLogger = logging.getLogger(name)
		thisInstance.__logger.setLevel(level.value)

		# 포맷 설정.
		if useShowProcessID:
			if useShowName:
				thisInstance.__formatter = Formatter("[%(asctime)s][%(name)s][%(process)d][%(levelname)s]%(message)s")
			else:
				thisInstance.__formatter = Formatter("[%(asctime)s][%(process)d][%(levelname)s]%(message)s")
		else:
			if useShowName:
				thisInstance.__formatter = Formatter("[%(asctime)s][%(name)s][%(process)d][%(levelname)s]%(message)s")
			else:
				thisInstance.__formatter = Formatter("[%(asctime)s][%(process)d][%(levelname)s]%(message)s")

		# 콘솔 로그.
		thisInstance.__consoleHandler = StreamHandler()
		thisInstance.__consoleHandler.setLevel(level.value)
		thisInstance.__consoleHandler.setFormatter(thisInstance.__formatter)
		thisInstance.__logger.addHandler(thisInstance.__consoleHandler) # 큐 핸들러 사용시 제외.

		# 파일 로그.
		if filePath:
			thisInstance.__fileHandler = FileHandler(filePath)
			thisInstance.__fileHandler.setLevel(level.value)
			thisInstance.__fileHandler.setFormatter(thisInstance.__formatter)
			thisInstance.__logger.addHandler(thisInstance.__fileHandler) # 큐 핸들러 사용시 제외.

		# # 큐 로그.
		# thisInstance.__queue = Queue()
		# thisInstance.__queueHandler = QueueHandler(thisInstance.__queue)
		# thisInstance.__logger.addHandler(thisInstance.__queueHandler)

		# # 큐 리스너.
		# handlers = list()
		# handlers.append(thisInstance.__consoleHandler)
		# if filePath:
		# 	handlers.append(thisInstance.__fileHandler)
		# thisInstance.__queueListener = QueueListener(thisInstance.__queue, *handlers)
		# thisInstance.__queueListener.start()


	#--------------------------------------------------------------------------------
	# 기록.
	#--------------------------------------------------------------------------------
	def __Log(thisInstance, level: LogLevel, message: object) -> None:
		if isinstance(message, str):
			if not message.startswith("["):
				message = f" {message}"
		else:
			text = str(message)
			if not text.startswith("["):
				message = f" {text}"

		thisInstance.__logger.log(level.value, message)


	#--------------------------------------------------------------------------------
	# 기록.
	#--------------------------------------------------------------------------------
	def LogDebug(thisInstance, message: object) -> None:
		thisInstance.__Log(LogLevel.DEBUG, message)


	#--------------------------------------------------------------------------------
	# 기록.
	#--------------------------------------------------------------------------------
	def LogInfo(thisInstance, message: object) -> None:
		thisInstance.__Log(LogLevel.INFO, message)


	#--------------------------------------------------------------------------------
	# 기록.
	#--------------------------------------------------------------------------------
	def LogWarning(thisInstance, message: object) -> None:
		thisInstance.__Log(LogLevel.WARNING, message)


	#--------------------------------------------------------------------------------
	# 기록.
	#--------------------------------------------------------------------------------
	def LogError(thisInstance, message: object) -> None:
		thisInstance.__Log(LogLevel.ERROR, message)


	#--------------------------------------------------------------------------------
	# 기록.
	#--------------------------------------------------------------------------------
	def LogException(thisInstance, message: object, *argumentTuple) -> None:
		thisInstance.__logger.exception(message, *argumentTuple, True, True, 8)


	#--------------------------------------------------------------------------------
	# 기록.
	#--------------------------------------------------------------------------------
	def LogCritical(thisInstance, message: object) -> None:
		thisInstance.__Log(LogLevel.CRITICAL, message)


#--------------------------------------------------------------------------------
# 참조 모듈 목록.
#--------------------------------------------------------------------------------
from __future__ import annotations
from typing import Awaitable, Callable, Final, Generic, Iterable, Iterator, Optional, Sequence, Type, TypeVar, Tuple, Union
from typing import ItemsView, KeysView, ValuesView
from typing import IO, TextIO, BinaryIO
from typing import Any, List, Dict, Set
from typing import cast, overload
import builtins
import os
import setuptools


#--------------------------------------------------------------------------------
# 전역 상수 목록.
#--------------------------------------------------------------------------------
UTF8: str = "utf-8"
READ: str = "r"


#--------------------------------------------------------------------------------
# 참조 메타 데이터 목록.
#--------------------------------------------------------------------------------
NAME: str = "xid-xpl"
# VERSION: str = "0.0.2" # 접두어 v 붙여도 알아서 정규화하면서 제거됨.
AUTHOR: str = "xidware"
AUTHOR_EMAIL: str = "developer@xidware.com"
DESCRIPTION: str = "Xidware Python Library"
LONG_DESCRIPTION_CONTENT_TYPE: str = "text/markdown"
URL: str = "https://xidware.com"
PYTHON_REQUIRES: str = ">=3.9.7" # 실제로는 3.10.11 이상.
LONGDESCRIPTION: str = str()
with open(file = "README.md", mode = READ, encoding = UTF8) as file: LONGDESCRIPTION = file.read()
with open(file = "VERSION", mode = READ, encoding = UTF8) as file: VERSION = file.read()


#--------------------------------------------------------------------------------
# 빌드.
#--------------------------------------------------------------------------------
setuptools.setup(
	name = NAME,
	version = VERSION,
	author = AUTHOR,
	author_email = AUTHOR_EMAIL,
	description = DESCRIPTION,
	long_description = LONGDESCRIPTION,
	long_description_content_type = LONG_DESCRIPTION_CONTENT_TYPE,
	url = URL,
	packages = setuptools.find_packages(where = "src"),
	include_package_data = True,
	package_dir = { "": "src" },
	package_data = {
		"": [
			"res/*"
		],
	},
	scripts = [

	],
	entry_points = {
		"console_scripts": [
		]
	},
	install_requires = [
	],
	classifiers = [
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent"
	],
	python_requires = PYTHON_REQUIRES
)


#--------------------------------------------------------------------------------
# 참조 모듈 목록.
#--------------------------------------------------------------------------------
from __future__ import annotations
from typing import Awaitable, Callable, Final, Generic, Iterable, Iterator, Optional, Sequence, Type, TypeVar, Tuple, Union
from typing import ItemsView, KeysView, ValuesView
from typing import IO, TextIO, BinaryIO
from typing import Any, List, Dict, Set
from typing import cast, overload
import builtins
from ..core import Reflection, BaseClass
# from threading import Lock
# from asyncio import Lock


#--------------------------------------------------------------------------------
# 기본 저장소 클래스.
# - 저장소 클래스를 통해 객체를 호출하면 최초 1회 할당 이후 계속 같은 객체를 반환한다.
# - 저장소 클래스 내부에서 생성되는 탓에 생성시 인자가 없는 객체여야 한다.
# - 만일 생성시 인자가 들어간다면 최초 생성인지 사용자가 판단해서 인자를 넣어주어야 한다.
# - 혹은 Link() 를 통해 외부에서 할당한 인스턴스를 넣어둘 수 있다.
# - 이 클래스를 상속받아서 독자적인 저장소 클래스를 만들 수 있다.
#--------------------------------------------------------------------------------
T = TypeVar("T", bound = Any)
class BaseRepository(BaseClass):
	#--------------------------------------------------------------------------------
	# 저장소에 들어가는 항목.
	#--------------------------------------------------------------------------------
	class RepositoryData(BaseClass):
		#--------------------------------------------------------------------------------
		# 멤버 변수 목록.
		#--------------------------------------------------------------------------------
		type: type
		instance: object


		#--------------------------------------------------------------------------------
		# 초기화 라이프사이클 메서드.
		#--------------------------------------------------------------------------------
		def __init__(thisInstance) -> None:
			base = super()
			base.__init__()
			thisInstance.type = None
			thisInstance.instance = None


	#--------------------------------------------------------------------------------
	# 클래스 멤버 변수 목록.
	#--------------------------------------------------------------------------------
	__IsInitialized: bool = False
	__TargetInstances: Dict[type, RepositoryData] = dict()


	#--------------------------------------------------------------------------------
	# 이벤트 체크.
	#--------------------------------------------------------------------------------
	@classmethod
	def CheckEvent(thisClassType) -> None:
		if not thisClassType.__IsInitialized:
			thisClassType.OnInitialize()
			thisClassType.__IsInitialized = True


	#--------------------------------------------------------------------------------
	# 초기화 라이프사이클 메서드됨.
	#--------------------------------------------------------------------------------
	@classmethod
	def OnInitialize(thisClassType) -> None:
		pass


	#--------------------------------------------------------------------------------
	# 파괴 라이프사이클 메서드.
	#--------------------------------------------------------------------------------
	@classmethod
	def OnFinalize(thisClassType) -> None:
		thisClassType.__TargetInstances.clear()


	#--------------------------------------------------------------------------------
	# 저장소에 모든 인스턴스 제거.
	#--------------------------------------------------------------------------------
	@classmethod
	def Cleanup(thisClassType) -> None:
		thisClassType.CheckEvent()
		thisClassType.__TargetInstances.clear()
	

	#--------------------------------------------------------------------------------
	# 저장소에 기존 인스턴스 설정 (기존 것이 있다면 제거 후 설정).
	#--------------------------------------------------------------------------------
	@classmethod
	def Link(thisClassType, instance: T) -> bool:
		thisClassType.CheckEvent()
		if not instance:
			return False
		
		isType: bool = Reflection.IsInstanceType(instance, type)
		if isType:
			return False

		instanceClassType: Type[T] = instance.__class__
		if instanceClassType in thisClassType.__TargetInstances:
			del thisClassType.__TargetInstances[instanceClassType]
		thisClassType.__TargetInstances[instanceClassType] = instance
		return True
	

	#--------------------------------------------------------------------------------
	# 저장소에 신규 인스턴스 생성 (기존 것이 있다면 제거 후 생성).
	# - 생성자에 값을 넣기 위해 argumentTuple과 argumentDictionary에 값을 넣어주어야 한다.
	# - 물론 값을 넣지 않아도 상관없으며 그 경우 인자 없는 생성자가 호출된다. (생성자에 인자가 존재한다면 주의)
	#--------------------------------------------------------------------------------
	@classmethod
	def Set(thisClassType, targetClassType: Type[T], *argumentTuple: Any, **argumentDictionary: Any) -> T:
		thisClassType.CheckEvent()
		if thisClassType.Contains(targetClassType):
			del thisClassType.__TargetInstances[targetClassType]

		item = BaseRepository.RepositoryData()
		item.type = targetClassType
		item.instance = targetClassType(*argumentTuple, **argumentDictionary)
		thisClassType.__TargetInstances[targetClassType] = item
		return item.instance
	

	#--------------------------------------------------------------------------------
	# 저장소에 존재하는 인스턴스 반환.
	# - 없다면 신규 생성하며 이 때 생성자에 값을 넣기 위해 args와 kwargs 값을 넣어주어야 한다.
	# - 물론 값을 넣지 않아도 상관없으며 그 경우 인자 없는 생성자가 호출된다. 생성자에 인자가 존재한다면 주의)
	#--------------------------------------------------------------------------------
	@classmethod
	def Get(thisClassType, targetClassType: Type[T], *argumentTuple: Any, **argumentDictionary: Any) -> T:
		thisClassType.CheckEvent()
		if thisClassType.Contains(targetClassType):
			item: BaseRepository.RepositoryData = thisClassType.__TargetInstances[targetClassType]
			return item.instance
		else:
			item: BaseRepository.RepositoryData = BaseRepository.RepositoryData()
			item.type = targetClassType
			item.instance = targetClassType(*argumentTuple, **argumentDictionary)
			thisClassType.__TargetInstances[targetClassType] = item
			return item.instance

	#--------------------------------------------------------------------------------
	# 저장소에 해당 인스턴스가 존재하는지 여부.
	#--------------------------------------------------------------------------------
	@classmethod
	def Contains(thisClassType, targetClassType: Type[T]) -> bool:
		thisClassType.CheckEvent()
		if targetClassType not in thisClassType.__TargetInstances:
			return False
		return True
	

	#--------------------------------------------------------------------------------
	# 저장소에 존재하는 인스턴스 반환.
	# - Get과 동일하나 부모를 인자로 넣으면 자식이 있을 경우 그 자식을 가져온다.
	# - 따라서 없으면 자동으로 대상을 추가 할 수가 없어서 안에 없다면 None을 뱉는다.
	#--------------------------------------------------------------------------------
	@classmethod
	def GetInherited(thisClassType, targetClassType: Type[T]) -> T:
		thisClassType.CheckEvent()
		if thisClassType.Contains(targetClassType):
			item: BaseRepository.RepositoryData = thisClassType.__TargetInstances[targetClassType]
			return item.instance
		else:
			for instanceClassType in thisClassType.__TargetInstances:
				if not issubclass(instanceClassType, targetClassType): # instanceClassType의 부모가 targetClassType일 때 참.
					continue
				item: BaseRepository.RepositoryData = thisClassType.__TargetInstances[instanceClassType]
				return item.instance
			return None